<?php

namespace Intergo\BaseClient;

use GuzzleHttp\RequestOptions;
use Intergo\BaseClient\Enums\HttpMethodEnum;
use Intergo\BaseClient\Requests\AuthenticationRequest;
use Intergo\BaseClient\Requests\BaseRequest;
use Intergo\BaseClient\Requests\RequestInterface;
use Intergo\BaseClient\Responses\BaseResponse;
use Intergo\BaseClient\Responses\PaginationResponse;
use Intergo\BaseClient\Responses\ErrorResponse;
use Intergo\BaseClient\Responses\BaseResponseInterface;
use Psr\Http\Message\ResponseInterface;
use Throwable;

abstract class Client
{
    protected ApiHandler $apiHandler;

    protected string $baseUrl;
    protected string $traceId;

    /**
     * @param string $baseUrl
     * @param $traceId
     */
    public function __construct(string $baseUrl, $traceId)
    {
        $this->baseUrl = $baseUrl;
        $this->traceId = $traceId;
        $this->setApiHandler();
    }

    /**
     * @return void
     */
    abstract protected function setApiHandler();

    /**
     * @param $endpoint
     * @param mixed $urlParams
     * @param mixed $bodyParams
     * @return BaseResponseInterface
     */
    public function execute($endpoint, $urlParams = [], $bodyParams = [])
    {
        $request = new BaseRequest($this->getUrl($endpoint, $urlParams), $this->traceId);
        return $this->process($request, $bodyParams, $endpoint);
    }

    /**
     * @param $endpoint
     * @param $bearerToken
     * @param mixed $urlParams
     * @param mixed $bodyParams
     * @return BaseResponseInterface
     */
    public function executeAuth($endpoint, $bearerToken, $urlParams = [], $bodyParams = [])
    {
        $request = new AuthenticationRequest($this->getUrl($endpoint, $urlParams), $bearerToken, $this->traceId);
        return $this->process($request, $bodyParams, $endpoint);
    }

    /**
     * @param string $endpoint
     * @param string $method
     * @param mixed $urlParams
     * @param mixed $bodyParams
     * @param ?string $bearerToken
     * @return BaseResponseInterface
     */
    public function executeCustom(
        string $endpoint,
        string $method = HttpMethodEnum::POST,
        $urlParams = [],
        $bodyParams = [],
        ?string $bearerToken = null
    )
    {
        $url = $this->apiHandler->prepareParameterizedURL($endpoint, $urlParams);
        $request = $bearerToken ? new AuthenticationRequest($url, $this->traceId) : new BaseRequest($url, $this->traceId);
        return $this->processRequest($request, $bodyParams, $method);
    }


    /**
     * @param RequestInterface $request
     * @param $bodyParams
     * @param $endpoint
     * @return BaseResponseInterface
     */
    protected function process(RequestInterface $request, $bodyParams, $endpoint)
    {
        return $this->processRequest($request, $bodyParams, $this->apiHandler::getMethod($endpoint));
    }

    /**
     * @param $endpoint
     * @param $urlParams
     * @return string
     */
    protected function getUrl($endpoint, $urlParams)
    {
        return $this->baseUrl . $this->apiHandler::getUrl($endpoint, $urlParams);
    }

    /**
     * @param RequestInterface $request
     * @param array $params
     * @param string $method
     * @return BaseResponseInterface
     */
    protected function processRequest(RequestInterface $request, array $params, string $method = HttpMethodEnum::POST)
    {
        try {
            $response = $request->execute($method, [
                $this->requestOptions($method) => $this->removeEmptyValues($params)
            ]);
        } catch (Throwable $exception) {
            return $this->getErrorResponse($exception);
        }

        if (isset($response['current_page']) && isset($response['total'])) {
            return new PaginationResponse($response);
        }

        return $this->getBaseResponse($response);
    }

    /**
     * @param string $method
     * @return string
     */
    protected function requestOptions(string $method)
    {
        return $method === HttpMethodEnum::GET ? RequestOptions::QUERY : RequestOptions::JSON;
    }

    /**
     * @param $params
     * @return array
     */
    protected function removeEmptyValues($params)
    {
        return array_filter($params, function ($a) {
            return $a !== null;
        });
    }

    /**
     * @param $exception
     * @return BaseResponseInterface
     */
    protected function getErrorResponse($exception)
    {
        return new ErrorResponse($exception);
    }

    /**
     * @param $response
     * @return BaseResponseInterface
     */
    protected function getBaseResponse($response)
    {
        return new BaseResponse($response);
    }
}
